package joturlsdk;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.TimeZone;

import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import javax.ws.rs.client.Client;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.client.Invocation.Builder;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import org.json.JSONObject;

import io.dropwizard.jersey.gzip.GZipDecoder;

public class JotUrlSDK {

	public static final String BASE_URL = "https://joturl.com/a/i1/";

	private String username;
	private String public_key;
	private String private_key;
	private String base_url;
	private String session_id;

	/**
	 * Creates an instance of the JotUrl SDK.
	 *
	 * @param username    the username used to login into JotURL dashboard
	 * @param public_key  public api key, you can find it on
	 *                    https://www.joturl.com/reserved/settings.html#tools-api
	 * @param private_key private api key, you can find it on
	 *                    https://www.joturl.com/reserved/settings.html#tools-api
	 * 
	 */
	public JotUrlSDK(String username, String public_key, String private_key) {
		this.username = username;
		this.public_key = public_key;
		this.private_key = private_key;
		this.base_url = BASE_URL;
		this.session_id = "";
	}

	public void setBaseURL(String base_url) {
		this.base_url = base_url;
	}

	public String create_sha256(String key, String message) throws NoSuchAlgorithmException, InvalidKeyException {
		Mac sha256_HMAC = Mac.getInstance("HmacSHA256");
		SecretKeySpec secret_key = new SecretKeySpec(private_key.getBytes(), "HmacSHA256");
		sha256_HMAC.init(secret_key);

		return hex(sha256_HMAC.doFinal(message.getBytes()));
	}

	public String getHash(String value, String time) throws NoSuchAlgorithmException, InvalidKeyException {
		String msg = value + ":" + time;

		return create_sha256(private_key, msg);
	}

	public String getHash(String value) throws NoSuchAlgorithmException, InvalidKeyException {
		return getHash(value, gmdate());
	}

	public String buildQuery(Map<String, String> params) {
		StringBuilder sb = new StringBuilder();

		for (Map.Entry<String, String> entry : params.entrySet()) {
			if (sb.length() > 0) {
				sb.append('&');
			}

			sb.append(encodeParam(entry.getKey()));
			sb.append("=");
			sb.append(encodeParam(entry.getValue()));
		}

		return sb.toString();
	}

	public URI buildURL(String endpoint, Map<String, String> params)
			throws NoSuchAlgorithmException, InvalidKeyException, URISyntaxException {
		if (params == null) {
			params = new HashMap<String, String>();
		}

		String parameters = buildQuery(merge(params, getSecurityParameters()));

		return new URI(base_url + endpoint + (!isEmpty(parameters) ? '?' + parameters : ""));
	}

	public Map<String, String> getSecurityParameters() throws NoSuchAlgorithmException, InvalidKeyException {
		HashMap<String, String> parameters = new HashMap<String, String>();

		if (!isEmpty(session_id)) {
			parameters.put("_sid", session_id);
			parameters.put("_h", getHash(session_id));
		}

		return parameters;
	}

	private JSONObject download(URI uri, Map<String, String> postParameters) {
		Client client = ClientBuilder.newBuilder().register(new GZipDecoder()).build();

		WebTarget target = client.target(uri);

		Builder request = target.request();
		request.accept(MediaType.APPLICATION_JSON);

		Response response = null;

		if (postParameters != null) {
			String payload = buildQuery(postParameters);
			response = request.post(Entity.entity(payload, MediaType.APPLICATION_JSON));
		} else {
			response = request.get();
		}

		if (response == null || (int) (response.getStatus() / 100) != 2) {
			throw new RuntimeException("Failed : HTTP error code : " + response.getStatus());
		}

		JSONObject json = new JSONObject(response.readEntity(String.class));

		JSONObject status = json.has("status") ? json.getJSONObject("status") : null;
		if (status == null || !status.has("code")) {
			throw new RuntimeException("Invalid status");
		}

		if ((int) (status.getInt("code") / 100) != 2) {
			String e = status.has("error") ? status.getString("error")
					: "Failed : HTTP error code : " + status.getInt("code");
			throw new RuntimeException(e);
		}

		JSONObject result = json.has("result") ? json.getJSONObject("result") : null;
		if (result == null) {
			throw new RuntimeException("Invalid result");
		}

		return result;
	}

	private JSONObject download(URI uri) {
		return download(uri, null);
	}

	public void login(Boolean force, JotUrlRunnable callback)
			throws InvalidKeyException, NoSuchAlgorithmException, URISyntaxException {
		if (!force && !isEmpty(session_id)) {
			callback.run();

			return;
		}

		String password = getHash(public_key);

		HashMap<String, String> parameters = new HashMap<String, String>();
		parameters.put("username", username);
		parameters.put("password", password);

		URI uri = buildURL("users/login", parameters);

		try {
			JSONObject r = download(uri);
			if (r != null && r.has("session_id")) {
				session_id = r.getString("session_id");

				callback.run();

				return;
			}
		} catch (RuntimeException e) {
			if (!force && e.getMessage().indexOf("INVALID Session") >= 0) {
				login(true, callback);
			}
			e.printStackTrace();

			throw new RuntimeException(e.getMessage());
		}

		throw new RuntimeException("Unable to login");
	}

	public void login(JotUrlRunnable callback)
			throws InvalidKeyException, NoSuchAlgorithmException, URISyntaxException {
		login(false, callback);
	}

	public void call(String endpoint, JotUrlRunnable callback, Map<String, String> parameters,
			Map<String, String> postParameters)
			throws InvalidKeyException, NoSuchAlgorithmException, URISyntaxException {
		login(new JotUrlRunnable() {
			public void run() {
				try {
					URI uri = buildURL(endpoint, parameters);

					JSONObject result = download(uri, postParameters);

					callback.run(result);
				} catch (InvalidKeyException | NoSuchAlgorithmException | URISyntaxException e) {
					throw new RuntimeException(e.getMessage());
				}
			}
		});
	}

	public void call(String endpoint, JotUrlRunnable callback)
			throws InvalidKeyException, NoSuchAlgorithmException, URISyntaxException {
		call(endpoint, callback, null, null);
	}

	public void callGET(String endpoint, Map<String, String> parameters, JotUrlRunnable callback)
			throws InvalidKeyException, NoSuchAlgorithmException, URISyntaxException {
		call(endpoint, callback, parameters, null);
	}

	public void callPOST(String endpoint, Map<String, String> postParameters, JotUrlRunnable callback)
			throws InvalidKeyException, NoSuchAlgorithmException, URISyntaxException {
		call(endpoint, callback, null, postParameters);
	}

	private static Map<String, String> merge(Map<String, String> src, Map<String, String> dst) {
		dst.forEach((key, value) -> src.put(key, value));

		return src;
	}

	private static boolean isEmpty(String s) {
		return s == null || s.length() == 0;
	}

	private static String hex(byte[] bytes) {
		StringBuilder result = new StringBuilder();
		for (byte aByte : bytes) {
			result.append(String.format("%02x", aByte));
		}

		return result.toString();
	}

	public static String gmdate() {
		String pattern = "yyy-MM-dd'T'HH:mm'Z'";
		SimpleDateFormat simpleDateFormat = new SimpleDateFormat(pattern);
		simpleDateFormat.setTimeZone(TimeZone.getTimeZone("UTC"));

		return simpleDateFormat.format(new Date());
	}

	private static String encodeParam(Object param) {
		String urlEncodeStr = "";
		try {
			urlEncodeStr = URLEncoder.encode(param.toString(), java.nio.charset.StandardCharsets.UTF_8.toString());
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}

		return urlEncodeStr;
	}
}